import os
from importlib import import_module
class ImproperlyConfigured(Exception): pass
import sys
APPS_MODULE_NAME = 'app'
import sys
from importlib.util import find_spec as importlib_find

class AppRegistry:
    _registry = {}
    
    def __init__(self) -> None:
        pass
    
    def get_registered_apps(self):
       return self._registry
   
    def register_app_class(self, cls, author=None, app_name=None, description=None, tags=(), **kwargs):
        app_dict = {
            'app_name': app_name,
            'cls': cls,
            'description': description,
            'tags': tags,
            'kwargs': kwargs
        }
        author_dict = self._registry.get(author, False)
        if not author_dict:
            self._registry[author] = [app_dict]
        else:
            # TODO: check if same name exists
            self._registry[author].append(app_dict)

    @staticmethod
    def import_balcony_apps(insalled_apps=None):
        for entry in insalled_apps:
            # import the entry
            try:
                app_module = import_module(entry)
                app_module
            except Exception as e:
                pass
            else:
                if module_has_submodule(app_module, APPS_MODULE_NAME):
                    mod_path = "%s.%s" % (entry, APPS_MODULE_NAME)
                    mod = import_module(mod_path)
                    mod
    # TODO: save the module info and return it
    
app_registry = AppRegistry()
class ResourceNodeRegistry:
    _registry = {}
    
    def register_class(self, cls, service_name=None, name=None):
        # TODO: check if service_name is correct get_available_service_names
        service_dict = self._registry.get(service_name, False)
        if not service_dict:
            self._registry[service_name] = {name: cls}
        else:
            custom_resource_node = service_dict.get(name, False)
            if custom_resource_node:
                raise Exception(
                    f"A custom ResourceNode is already registered with Service: {service_name}, Resource Node: {name}. Duplication is not allowed.")
            else:
                self._registry[service_name][name] = cls


    def _search_registry_for_service(cls, service_name):
        return cls._registry.get(service_name, {})

    def find_custom_class_for_resource_node(cls, service_name, node_name):
        services_custom_nodes = cls._search_registry_for_service(service_name)
        for custom_cls_name, _cls_obj in services_custom_nodes.items():
            # if compare_two_camel_case_words(custom_cls_name, node_name):
            if custom_cls_name.lower() == node_name.lower():
                return _cls_obj
        return False
    
    
    

def cached_import(module_path, class_name):
    # Check whether module is loaded and fully initialized.
    if not (
        (module := sys.modules.get(module_path))
        and (spec := getattr(module, "__spec__", None))
        and getattr(spec, "_initializing", False) is False
    ):
        module = import_module(module_path)
    return getattr(module, class_name)


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    try:
        return cached_import(module_path, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from err


def autodiscover_modules(*args, **kwargs):
    """
    Auto-discover INSTALLED_APPS modules and fail silently when
    not present. This forces an import on them to register any admin bits they
    may want.

    You may provide a register_to keyword parameter as a way to access a
    registry. This register_to object must have a _registry instance variable
    to access it.
    """

    # TODO static init
    register_to = kwargs.get("register_to")
    for app_config in app_registry.get_app_configs():
        for module_to_search in args:
            # Attempt to import the app's module.
            try:
                if register_to:
                    before_import_registry = copy.copy(register_to._registry)

                import_module("%s.%s" % (app_config.name, module_to_search))
            except Exception:
                # Reset the registry to the state before the last import
                # as this import will have to reoccur on the next request and
                # this could raise NotRegistered and AlreadyRegistered
                # exceptions (see #8245).
                if register_to:
                    register_to._registry = before_import_registry

                # Decide whether to bubble up this error. If the app just
                # doesn't have the module in question, we can ignore the error
                # attempting to import it, otherwise we want it to bubble up.
                if module_has_submodule(app_config.module, module_to_search):
                    raise


def module_has_submodule(package, module_name):
    """See if 'module' is in 'package'."""
    try:
        package_name = package.__name__
        package_path = package.__path__
    except AttributeError:
        # package isn't a package.
        return False

    full_module_name = package_name + "." + module_name
    try:
        return importlib_find(full_module_name, package_path) is not None
    except ModuleNotFoundError:
        # When module_name is an invalid dotted path, Python raises
        # ModuleNotFoundError.
        return False


def module_dir(module):
    """
    Find the name of the directory that contains a module, if possible.

    Raise ValueError otherwise, e.g. for namespace packages that are split
    over several directories.
    """
    # Convert to list because __path__ may not support indexing.
    paths = list(getattr(module, "__path__", []))
    if len(paths) == 1:
        return paths[0]
    else:
        filename = getattr(module, "__file__", None)
        if filename is not None:
            return os.path.dirname(filename)
    raise ValueError("Cannot determine directory containing %s" % module)



from re import finditer, compile
from typing import List
import inflect
inflect_engine = inflect.engine()  # used for same word comparing
import os
import boto3

_camel_case_regex_compiled = compile(
    r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)")

def _create_boto_session():
    profile_name = os.environ.get('AWS_PROFILE', False)
    region_name = os.environ.get('AWS_REGION', False)
    _kwargs_dict = {}
    if profile_name:
        _kwargs_dict['profile_name']=profile_name
    if region_name:
        _kwargs_dict['region_name']=region_name
    session = boto3.session.Session(**_kwargs_dict)
    return session 

def ifind_similar_names_in_list(search_for, search_in_list):
    # TODO: regex support
    result = [] 
    if not search_for:
        return result
    lower_search_for = search_for.lower()
    
    for l_name in search_in_list:
        if l_name.lower().startswith(lower_search_for):
            result.append(l_name)
    return result
# >>> items = set([-1, 0, 1, 2])
# >>> set([1, 2]).issubset(items)
# True
# >>> set([1, 3]).issubset(items)
# False

def str_relations(relations):
    return ", ".join([
        f"[{r.get('service_name')}.{r.get('operation_name')} > {r.get('target_path')}]"
        for r in relations
    ])

def get_all_available_services(session) -> List[str]:
    return session.get_available_services()

def compare_nouns(word1: str, word2: str):
    return inflect_engine.compare_nouns(word1, word2)

def compare_two_token_lists(token_list_one, token_list_two):
    if len(token_list_one) != len(token_list_two):
        return False
    token_by_token_match = [
        w1.lower()==w2.lower()
        for w1, w2 in zip(token_list_one, token_list_two)
    ]
    has_token_by_token_match = all(token_by_token_match)
    return has_token_by_token_match

def icompare_two_token_lists(token_list_one, token_list_two):
    if len(token_list_one) != len(token_list_two):
        return False
    token_by_token_match = [
        bool(inflect_engine.compare_nouns(w1.lower(), w2.lower()))
        for w1, w2 in zip(token_list_one, token_list_two)
    ]
    has_token_by_token_match = all(token_by_token_match)
    return has_token_by_token_match

def camel_case_split(identifier: str) -> List[str]:
    """Splits camel case string to it's tokens"""
    matches = finditer(
        _camel_case_regex_compiled, identifier
    )
    return [m.group(0) for m in matches]

def compare_two_camel_case_words(word1, word2):
    token_list_1 = [_.lower() for _ in camel_case_split(word1)]
    token_list_2 = [_.lower() for _ in camel_case_split(word2)]
    return compare_two_token_lists(token_list_1, token_list_2)


def icompare_two_camel_case_words(word1, word2):
    token_list_1 = [_.lower() for _ in camel_case_split(word1)]
    token_list_2 = [_.lower() for _ in camel_case_split(word2)]
    return icompare_two_token_lists(token_list_1, token_list_2)

def compare_two_tokens(token_one, token_two):
    return bool(inflect_engine.compare_nouns(token_one, token_two))


def get_last_token_of_target_path(target_path, delimiter='.'):
    split_return_key = target_path.split(delimiter)
    *prev_return_keys, last_return_key_name = split_return_key
    return last_return_key_name


try:
    from .utils import camel_case_split, compare_nouns, ifind_similar_names_in_list , icompare_two_token_lists, compare_two_camel_case_words, str_relations
    from .botocore_utils import *
    from .relations import RelationMap, FindRelationResultTypes
    from .reader import ServiceReader
    from .registries import ResourceNodeRegistry
    from .logs import get_logger, get_rich_console
    from .errors import Error
except ImportError:
    from utils import camel_case_split, compare_nouns, ifind_similar_names_in_list , icompare_two_token_lists, compare_two_camel_case_words, str_relations
    from botocore_utils import *
    from relations import RelationMap, FindRelationResultTypes
    from reader import ServiceReader
    from registries import ResourceNodeRegistry
    from logs import get_logger, get_rich_console
    from errors import Error

from typing import List, Set, Dict, Tuple, Optional, Union
from itertools import product as cartesian_product
import copy
from botocore.utils import ArgumentGenerator
from botocore.validate import validate_parameters
from botocore.hooks import EventAliaser
from botocore.model import OperationModel, ServiceModel
from rich.text import Text
from rich.panel import Panel
from rich.console import Group
from rich.padding import Padding
from rich.layout import Layout
import jmespath
import textwrap
from enum import Enum

logger = get_logger(__name__)

_resource_node_registry = ResourceNodeRegistry()
argument_generator = ArgumentGenerator()
console = get_rich_console()
class OperationType(str, Enum):
    GET = "get"
    GETS = "gets"
    LIST = "list"
    DESC = "desc"
    

class ResourceNode:
    def __init__(self, service_node: 'ServiceNode', name: str, operation_names: List[str]) -> None:
        self.service_node = service_node
        self.name = name
        self.operation_names = operation_names
        self._operation_models = {}
        

    def __init_subclass__(cls, service_name=None, name=None, **kwargs):
        super().__init_subclass__(**kwargs)
        _resource_node_registry.register_class(cls, service_name, name)

    def get_operation_types_and_names(self) -> Dict[OperationType, str]:
        operation_names = self.get_operation_names()
        types_to_operation_names = {}
        for op_name in operation_names:
            verb, *resource_node_name_tokens = camel_case_split(op_name)
            resource_node_name = ''.join(resource_node_name_tokens)
            if verb.lower() == 'describe':
                types_to_operation_names[OperationType.DESC.value] = op_name
            if verb.lower() == 'list':
                types_to_operation_names[OperationType.LIST.value] = op_name

            if verb.lower() == 'get':
                if self.name != resource_node_name:
                    types_to_operation_names[OperationType.GETS.value] = op_name
                else:
                    types_to_operation_names[OperationType.GET.value] = op_name
        return types_to_operation_names
        

    def get_operation_names(self) -> List[str]:
        """Returns the available operation names in the ResourceNode. 

        Returns:
            List[str]: Operation names tied to the ResourceNode
        """
        return self.operation_names

    # NOTE: +overrideable
    def define_extra_relations(self) -> List[Dict]:
        """Extra relations defined in the custom subclasses of ResourceNode

        Returns:
            List[Dict]: List of relations
        """
        return []
    
    # NOTE: +overrideable
    def get_operations_relations(self, operation_name:str) -> Tuple[Union[List[Dict], bool], Union[Error, None]]:
        """_summary_

        Args:
            operation_name (str): Name of the operation.
            relation_map (Optional[RelationMap], optional): RelationMap object for the Service Node. Defaults to None.

        Returns:
            Tuple[Union[List[Dict], bool], Union[Error, None]]: Returns value and error. 
                                                                True: No required parameters
                                                                False: Failure
                                                                List[Dict]: List of relations 
        """
        
        resource_node = self
        relation_map = self.service_node.get_relation_map()
        required_parameter_names = resource_node.get_required_parameter_names_from_operation_name(operation_name)
        required_parameter_names_to_relations_map = {
            r_param_name:None 
            for r_param_name in required_parameter_names
        }
        
        if not required_parameter_names:
            # no required parameters
            # logger.debug(f"NO REQUIRED PARAMETERS. [bold][blue]{self.service_node.name}[/].[green]{operation_name}[/][/] has no required parameters")
            return True, None

        operation_markup = f"[bold][green]{self.service_node.name}[/].[blue]{operation_name}[/][/]"
        req_param_markup = f"[bold magenta]{', '.join(required_parameter_names)}[/]"
        
        if len(required_parameter_names) == 1:
            # only one parameter exists
            selected_relations = None
            single_parameter_name = required_parameter_names[0]
            generated_relations_for_parameter = relation_map.get_parameters_generated_relations(single_parameter_name, operation_name)
            if not generated_relations_for_parameter:
                logger.debug(f"Failed to generate relations. {operation_markup} has a required parameter: {req_param_markup}")
                return False, Error("failed to generate relations", {'required_parameter_names': required_parameter_names, 'service':self.service_node.name, 'resource_node':self.name, 'operation_name':operation_name})
                # return False, FindRelationResultTypes.NoGeneratedParameters

            selected_relations = resource_node.find_best_relation_for_single_parameter(single_parameter_name, generated_relations_for_parameter)
            # required_parameter_names_to_relations_map[single_parameter_name]=selected_relations
            if selected_relations:
                logger.debug(f"[green]Success finding relations.[/] {operation_markup} has a required parameter: {req_param_markup}. Relation found: [yellow]{str_relations(selected_relations)}[/]")
                # return selected_relations, FindRelationResultTypes.RelationsFound
                return selected_relations, None
            else:
                logger.debug(f"Failed to choose the best relation for operation. {operation_markup} has required parameters: {req_param_markup}")
                return False, Error("failed to choose the best relation", {'service':self.service_node.name, 'resource_node':self.name, 'operation_name':operation_name, 'generated_relations_for_parameter':generated_relations_for_parameter})
                # return False, FindRelationResultTypes.CantDecideBetweenGeneratedParameters
        else:
            # multiple required parameters, zip and check
            # required_parameter_names = list(required_parameter_names_to_relations_map.keys())
            # temporary_parameters_to_relations_map = dict()
            for r_parameter_name in required_parameter_names:
                generated_relations_for_parameter = relation_map.get_parameters_generated_relations(r_parameter_name, operation_name)
                if generated_relations_for_parameter:
                    required_parameter_names_to_relations_map[r_parameter_name]=generated_relations_for_parameter
                # return False, FindRelationResultTypes.NoGeneratedParameters
                # selected_relations = resource_node.find_best_relation_for_single_parameter(r_parameter_name, generated_relations_for_parameter)
                # ignore relations to self

            _all_parameters_have_relations = [
                bool(_found_relations) 
                for _found_relations in required_parameter_names_to_relations_map.values()
                
            ]

            if not all(_all_parameters_have_relations):
                # all parameters doesnt have relations
                if any(_all_parameters_have_relations):
                    # some found but not all
                    partial_relations_str = ", ".join([
                        str_relations(_found_relations)
                        for _found_relations in required_parameter_names_to_relations_map.values()
                        if _found_relations
                    ])
                    logger.debug(f"NOT ALL RELATIONS FOUND. {operation_markup} has required parameters: {req_param_markup}. Partially found relations: {partial_relations_str}")
                    return False, Error("missing relations for parameter", {'service':self.service_node.name, 'resource_node':self.name, 'operation_name':operation_name, 'required_parameter_names_to_relations':required_parameter_names_to_relations_map})
                    # return False, FindRelationResultTypes.SomeRelationsFoundButNotAll
                else:
                    # nothing found
                    logger.debug(f"NO RELATIONS FOUND. {operation_markup} has required parameters: {req_param_markup}.")
                    # return False, FindRelationResultTypes.NoRelations
                    return False, Error("failed to generate relations", {'service':self.service_node.name, 'resource_node':self.name,'operation_name':operation_name,  'required_parameter_names_to_relations':required_parameter_names_to_relations_map})
            
            def check(list_of_relations):
                if not list_of_relations:
                    return False
                first_operation_name = list_of_relations[0].get('operation_name')
                return all([l_o_r.get('operation_name')==first_operation_name for l_o_r in list_of_relations])
       
            possible_relation_combinations = None
            relation_matrix = list(required_parameter_names_to_relations_map.values())
            if relation_matrix:
                relations_cartesian_product = cartesian_product(*relation_matrix)
                possible_relation_combinations = list(filter(check, relations_cartesian_product))            
            
            if not possible_relation_combinations:
                logger.debug(f"CAN'T DECIDE BTWN RELATIONS. {operation_markup} has required parameters: {req_param_markup}.")
                # return False, FindRelationResultTypes.CantDecideBetweenGeneratedParameters
                return False, Error("failed to choose the best relation", {'service':self.service_node.name, 'resource_node':self.name, 'operation_name':operation_name, 'generated_relations_for_parameter':generated_relations_for_parameter})
                
            
            # find common target_operation relations across the permutation.
            relation_chosen_from_possible_combinations = self.select_between_possible_relation_combinations_matrix(possible_relation_combinations)    
            if relation_chosen_from_possible_combinations:
                logger.debug(f"MULTIPLE PARAMETERS FOUND. {operation_markup} has req parameters: {req_param_markup}. Relations found: {str_relations(relation_chosen_from_possible_combinations)}")
                return relation_chosen_from_possible_combinations, None
            else:
                logger.debug(f"CAN'T DECIDE BTWN RELATIONS. {operation_markup} has req parameters: {req_param_markup}.")
                # return False, FindRelationResultTypes.CantDecideBetweenGeneratedParameters
                return False, Error("failed to choose the best relation", {'service':self.service_node.name, 'resource_node':self.name,'operation_name':operation_name, 'generated_relations_for_parameter':generated_relations_for_parameter})


    # NOTE: +overrideable
    def generate_jmespath_selector_from_relations(self, operation_name: str, relation_list: List[Dict]) -> str:
        """
        Tries to generate the jmespath selector string from given relations. Could be overridden with custom subclasses. 

        Args:
            operation_name (str): Name of the operation
            relation_list (List[Dict]): List of relations for the operation. Used to generate the jmespath selector.

        Returns:
            str: Jmespath selector string
        """
        #TODO: (a better way for generating multiple same level args.) https://kyverno.io/docs/writing-policies/jmespath/  $ yq e pod.yaml -o json | kyverno jp "spec.[initContainers, containers][]"
        _for_all_the_responses = '[*].'
        _flatten_two_times = '[][]'
        if not relation_list:
            return False
        # [*].Stacks[*].{StackName: StackName, age: age}[] -> [*].Stacks[*].StackName[][]
        # [*].Stacks[*].StackName[], [*].Stacks[*].Description[]  
        target_shape_and_min_nested_target_paths = None
        # TODO: if type is list, do something different!!
        if len(relation_list) == 1:
            relation = relation_list[0]
            selected_target_path = min(relation.get('target_path').split(','), key=lambda tp: tp.count('.'))
            target_shape_and_min_nested_target_paths = [
                (relation.get('target_shape_name'), selected_target_path)
            ]
        else:
            # multiple relations
            # all relations point to same operation_name
            target_shape_and_min_nested_target_paths = [    
                (
                    relation.get('target_shape_name'), 
                    min(
                        relation.get('target_path').split(','), 
                        key=lambda tp: tp.count('.')
                    )
                )
                for relation in relation_list
            ]
            # all_selected_target_paths_same_depth = False
            # if len(relations_and_min_nested_target_paths) >= 2:

            # ensure target paths are same deep nested
            first_target_shape_and_target_path, *rest_of_target_shape_and_target_path = target_shape_and_min_nested_target_paths
            _, first_target_path = first_target_shape_and_target_path
            first_target_path_deepness = first_target_path.count('.')
            first_target_path_before_last_attribute = first_target_path.rsplit('.', maxsplit=1)[0]
            all_selected_target_paths_same_depth = all([
                r_target_path.count('.') == first_target_path_deepness \
                    and r_target_path.rsplit('.', maxsplit=1)[0] == first_target_path_before_last_attribute
                for _, r_target_path in rest_of_target_shape_and_target_path
            ])
                    
            if not all_selected_target_paths_same_depth:
                logger.debug(f"[red]DOES NOT START WITH same json nest: {target_shape_and_min_nested_target_paths}")
                return False
                
        jmespath_nested_selector = None
        
        # if len(target_shape_and_min_nested_target_paths) == 1:
        #     jmespath_nested_selector
        # else:
        _jmespath_selectors_list = []
        before_last_attr = None
        for target_shape, target_path in target_shape_and_min_nested_target_paths:
            if '.' not in target_path:
                target_path = f".{target_path}"
            before_last_attr, last_attribute = target_path.rsplit('.', maxsplit=1)
                
            jmespath_nested_selector = f"{target_shape}: {last_attribute}"
            _jmespath_selectors_list.append(jmespath_nested_selector)
        if not before_last_attr:
             before_last_attr = ''
        jmespath_nested_selector = before_last_attr+'.{'+', '.join(_jmespath_selectors_list)+'}'
        jmespath_nested_selector
        flattened_jmespath_nested_selector = f"{_for_all_the_responses}{jmespath_nested_selector}{_flatten_two_times}"
        return flattened_jmespath_nested_selector

    # NOTE: +overrideable
    def _generate_raw_api_parameters_from_operation_data(self, operation_name:str, 
                                                         relations_of_operation:List[dict], 
                                                         related_operations_data: Union[Dict, List]) -> List:
        """Generates the jmespath selector and search the data with it. Output is the list of api_parameters dictionaries.
        Only considers the required parameters(raw parameters) of the operation. 
        > Note: Pagination parameters (e.g. `NextToken`, `MaxResults`) are not included. 

        Args:
            operation_name (str): Name of the operation
            relations_of_operation (List[dict]): Relations of the operation
            related_operations_data (Union[Dict, List]): All related operations data

        Returns:
            List: List of required _raw_ API parameters, not including pagination parameters
        """
        resource_node = self
        generated_jmespath_nested_selector = resource_node.generate_jmespath_selector_from_relations(operation_name, relations_of_operation)
        if not generated_jmespath_nested_selector:
            logger.debug(f"CAN'T GENERATE JMESPATH SELECTOR: {resource_node.name} {operation_name} {generated_jmespath_nested_selector}")
            return False, Error("failed to generate jmespath selector", {'service':self.service_node.name, 'resource_node':self.name,'operation_name':operation_name,})
        logger.debug(f"JMESPATH SELECTOR GENERATED: [blue]{generated_jmespath_nested_selector}[/], target operation: [bold][blue][{resource_node.name}[/].[green]{operation_name}[/]]")

        if not related_operations_data:
            logger.debug(f"NO OPERATION DATA FOUND. {related_operations_data=}")

        # FIXME
        # the first relation we will get 
        direct_relation = relations_of_operation[0]
        direct_related_operation = direct_relation.get('operation_name')
        directly_related_operation_data = related_operations_data.get(direct_related_operation)
        raw_api_parameters_list = jmespath.search(generated_jmespath_nested_selector, directly_related_operation_data)
        
        if raw_api_parameters_list == []:
            # successfull jmespath search that yielded no results. operation data might be empty
            return raw_api_parameters_list, Error("related resources not found", {'service':self.service_node.name, 'resource_node':self.name,'operation_name':operation_name,})
        elif not raw_api_parameters_list:
            logger.debug(f"CANT GENERATE API PARAMETERS LIST WITH [bold][red]{generated_jmespath_nested_selector}[/] {related_operations_data=}")
            return False, Error("failed to generate api parameters", {'service':self.service_node.name, 'resource_node':self.name,'operation_name':operation_name,})
        
        return raw_api_parameters_list, None
    
    # NOTE: +overrideable
    def create_valid_api_parameters_list(self, operation_name:str,
                                         related_operations_data: Union[List, Dict], 
                                         relations_of_operation:List[Dict], 
                                         raw_api_parameters_list:List) -> List:
        """Uses the raw API parameters list and appends pagination and other kind of parameters to them.

        Args:
            operation_name (str): Name of the operation.
            related_operations_data (Union[List, Dict]): All related operations data
            relations_of_operation (List[Dict]): Relations of the operation
            raw_api_parameters_list (List): Generated raw API parameters 

        Returns:
            List: Valid API parameters to call the boto operation with
        """
        
        operation_model = self.get_operation_model(operation_name)
        required_parameter_names = self.get_required_parameter_names_from_operation_name(operation_name)

        input_shape = get_input_shape(operation_model)
        generated = {}
        
        if input_shape:
            generated = argument_generator.generate_skeleton(input_shape)
        
        result = {}

        for r_parameter_name in required_parameter_names:
            r_key = find_key_in_dict_keys(r_parameter_name, generated)
            if r_key:
                result[r_key] = generated[r_key]
                
        # handle MaxResults
        max_results_value = get_max_results_value_from_shape(input_shape)
        max_results_key = find_key_in_dict_keys('maxresults', generated)
        if max_results_key and max_results_value:
            result[max_results_key]=max_results_value
        # sometimes MaxResults can be seen as non required parameter, but in fact is
        if not required_parameter_names:
            return [result]
        
        
        search_to_target_names = {relation.get('target_shape_name'):relation.get('search_shape_name') for relation in relations_of_operation}
        
        generated_api_params = []
        if raw_api_parameters_list == False:
            # TODO: add logic here
            logger.debug(f"FAILED TO CREATE VALID API PARAMETERS. Required Parameters are: [bold]{required_parameter_names}[/]")
            return False
        for raw_api_parameter_dict in raw_api_parameters_list:
            result_copy = copy.deepcopy(result)
            for raw_key in raw_api_parameter_dict.keys():
                found_parameter_key = find_key_in_dict_keys(raw_key, result_copy)
                if not found_parameter_key:
                    # search to target name
                    new_key = find_key_in_dict_keys(raw_key, search_to_target_names)
                    if new_key:
                        found_parameter_key = find_key_in_dict_keys(search_to_target_names[new_key], result_copy)
                result_copy[found_parameter_key] = raw_api_parameter_dict[raw_key]
            generated_api_params.append(result_copy)
        # Handle NextToken and DryRun
        # dry_run_key_name = find_key_in_dict_keys('dryrun', generated)
        # generated.pop(dry_run_key_name)
        # next_token_key_name = find_key_in_dict_keys('nexttoken', generated)
        # generated.pop(next_token_key_name)
        # generated
        # # handle multiple required parameters
        # generated_api_params_list = []
        # for r_parameter_name in required_parameter_names:
        #     r_parameter_name_key = find_key_in_dict_keys(r_parameter_name, generated)
        #     r_parameter_value = generated.get(r_parameter_name_key)
        #     # if type
            
        #     # result[r_parameter_name_key] = ?
        # generated
        return generated_api_params
    
    # NOTE: +overrideable
    def generate_api_parameters_from_operation_data(self, operation_name:str, 
                                                    relations_of_operation:List[Dict], 
                                                    related_operations_data: Union[List, Dict]) -> Tuple[Union[List, bool], Union[Error, None]]:
        """Generates API parameters for the given operation including pagination parameters.

        Args:
            operation_name (str): Name of the operation
            relations_of_operation (List[Dict]): Relations of the operation
            related_operations_data (Union[List, Dict]): All related operations data

        Returns:
            List: _description_
        """
        resource_node = self
        resource_node_name = resource_node.name
        no_required_parameters = relations_of_operation == []

        if no_required_parameters:
            # Even though we know there are no required parameters, 
            # some parameters like MaxResults must be filled.
            api_parameters = resource_node.create_valid_api_parameters_list(operation_name, related_operations_data, [], [])
            return api_parameters, None

        related_operations_data
        raw_api_parameters_list, raw_param_error = self._generate_raw_api_parameters_from_operation_data(operation_name, relations_of_operation, related_operations_data)
        if raw_param_error is not None:
            # failed to generate raw api parameters list 
            return False, raw_param_error

        api_parameters_list = resource_node.create_valid_api_parameters_list(operation_name, related_operations_data, relations_of_operation, raw_api_parameters_list)
        if not api_parameters_list:
            return api_parameters_list, Error("failed to generate api parameters", {'service':self.service_node.name, 'resource_node':self.name,'operation_name':operation_name})
        
        return api_parameters_list, None
    
    def print_operation(self, operation_name:str) -> None:
        operation_panel = self._rich_operation_details_panel(operation_name)
        console.print(operation_panel)

    def _rich_operation_details_panel(self, operation_name:str) -> Panel:
        
        operation_model = self.get_operation_model(operation_name)
        input_shape = get_input_shape(operation_model)
        output_shape = operation_model.output_shape
        
        input_shape_tree = Text("[yellow]No Input Shape Found")
        if input_shape:
            input_shape_tree = generate_rich_tree_from_shape(input_shape)
        output_shape_tree = generate_rich_tree_from_shape(output_shape)
        
        operation_docs = cleanhtml(operation_model.documentation)
        panel_group = Group(
            Padding(f"[bold underline]Documentation:[/] {operation_docs}", (1, 2)),
            Panel(input_shape_tree, title=f'Input: [yellow]{input_shape.name}', title_align='left', padding=(1,1)),
            Panel(output_shape_tree, title=f'Output: [yellow]{output_shape.name}', title_align='left', padding=(1,1)),
            
        )
        Group()
        
        required_parameters = self.get_required_parameter_names_from_operation_name(operation_name)
        _title = f"[bold]Operation[/]: [magenta bold]{operation_name}[/]"
        if required_parameters:
            _title = f"{_title}   [white][Required: {', '.join(required_parameters)}]"
        operation_panel = Panel(panel_group, title=_title, highlight=True, title_align='left')
        return operation_panel

    def select_between_possible_relation_combinations_matrix(self, possible_relation_combinations: List[List[Dict]])->List[Dict]:
        """For given relation combinations matrix, select the best possible combination of relations.
        Gives scores to relation combinations, selects the one with the highest score.

        Args:
            possible_relation_combinations (List[List[Dict]]): Matrix of all possible combinations of multiple relations

        Returns:
            List[Dict]: Selected relations list
        """
        if len(possible_relation_combinations)==1:
            return possible_relation_combinations[0]
        # each combination will have the same operation name
        # prefer using non-GET function
        # prefer the operation that has no required parameters
        combination_index_to_score_map = {str(i):0 for i in range(len(possible_relation_combinations))}
        
        for i, relation_combination in enumerate(possible_relation_combinations):
            current_combination_point = 0
            combinations_operation_name = relation_combination[0].get('operation_name') # promised to all have same operation
            
            # negative point if it has required parameters
            combinations_required_parameters = self.get_required_parameter_names_from_operation_name(combinations_operation_name)
            current_combination_point -= len(combinations_required_parameters)
            
            # negative point if its a get function, we'd like to prefer list or describe functions
            # if combinations_operation_name.lower().startswith('get'):
            #     current_combination_point -= 1
            
            combination_index_to_score_map[str(i)] = current_combination_point
        selected_index_str, selected_score = max(combination_index_to_score_map.items(), key=lambda index_and_score: index_and_score[1])
        selected_combination = possible_relation_combinations[int(selected_index_str)]
        
        return selected_combination

    def select_between_possible_relation_combinations_list(self, possible_relation_list):
        if len(possible_relation_list)<1:
            return False
        if len(possible_relation_list)==1:
            return possible_relation_list[0]
        # each combination will have the same operation name
        # prefer using non-GET function
        # prefer the operation that has no required parameters
        combination_index_to_score_map = {str(i):0 for i in range(len(possible_relation_list))}
        
        get_operation_count  = [p_rel.get('operation_name').startswith('Get') for p_rel in possible_relation_list].count(True)
        for i, relation_dict in enumerate(possible_relation_list):
            current_combination_point = 0
            relations_operation_name = relation_dict.get('operation_name') # promised to all have same operation
            
            # negative point if it has required parameters
            combinations_required_parameters = self.get_required_parameter_names_from_operation_name(relations_operation_name)
            current_combination_point -= len(combinations_required_parameters)
            
            # todo: add if not plural!
            # negative point if its a get function, we'd like to prefer list or describe functions
            if get_operation_count <= 1 and relations_operation_name.lower().startswith('get'):
                current_combination_point -= 1
            
            combination_index_to_score_map[str(i)] = current_combination_point
        selected_index_str, selected_score = max(combination_index_to_score_map.items(), key=lambda index_and_score: index_and_score[1])
        selected_combination = possible_relation_list[int(selected_index_str)]
        
        return selected_combination
    
    def find_best_relation_for_single_parameter(self, parameter_name:str, relation_list:List[Dict])-> List[Dict]:
        _parameter_name_tokens = camel_case_split(parameter_name)
        non_id_parameter_tokens = [
            p_token for p_token in _parameter_name_tokens
            if p_token.lower() not in IDENTIFIER_NAMES
        ]
        
        # if not non_id_parameter_tokens:
        #     # only contains id 
        #     non_id_parameter_tokens  =_parameter_name_tokens

        same_resource_name_relations = []
        for relation_dict in relation_list:
            operation_name = relation_dict.get('operation_name')
            operation_verb, *operation_name_tokens = camel_case_split(operation_name)
            # relation_dict.update({'operation_verb': operation_verb, 'operation_name_tokens': operation_name_tokens})
            # TODO: here
            all_tokens_match = icompare_two_token_lists(
                non_id_parameter_tokens, operation_name_tokens)
            if all_tokens_match:
                same_resource_name_relations.append(relation_dict)
                    
        # what we are left with is different operations names 
        selected_relation = self.select_between_possible_relation_combinations_list(same_resource_name_relations)
        if selected_relation:
            return [selected_relation]
        return False
    
    # TODO: memoization
    def get_required_parameter_names_from_operation_name(self, operation_name:str)-> List[str]:
        
        operation_model = self.get_operation_model(operation_name)
        input_shape = get_input_shape(operation_model)
        if not input_shape:
            return []
        required_parameter_names = getattr(input_shape, 'required_members', False)
        max_results_key = find_key_in_dict_keys('maxresults', required_parameter_names)
        if max_results_key:
            required_parameter_names.remove(max_results_key)
        return required_parameter_names
    
    def get_all_required_parameter_names(self)-> List[str]:
        all_required_names = []
        for operation_name in self.operation_names:
            required_shapes = self.get_required_parameter_names_from_operation_name(
                operation_name)
            all_required_names.extend(required_shapes)
        return all_required_names
    
    def get_operation_model(self, operation_name: str) -> OperationModel:
        if operation_name in self._operation_models:
            return self._operation_models[operation_name]
        service_model = self.service_node.get_service_model()
        operation_model = service_model.operation_model(operation_name)
        self._operation_models[operation_name] = operation_model
        return operation_model

    def json(self) -> Dict:
        return {
            "service_node_name": self.service_node.name,
            "name": self.name,
            "operation_names": self.operation_names
        }

    def __str__(self) -> str:
        # return f"[{self.service_node.name}.{self.name}]"
        return f"[{self.name}]"

    def __rich__(self) -> str:
        return 'ss'
 
 


class ServiceNode:
    def __init__(self, name, session):
        self.name = name
        self.session = session
        self.client = self.session.client(self.name)
        self.resource_nodes = None
        self._relation_map = None
        self._reader = None
        self._read_operation_name_to_tokens_map = None


    def read(self, resource_node_name):
        reader = self.get_service_reader()
        return reader.read_resource_node(resource_node_name)
    
    def get_service_reader(self) -> ServiceReader:
        """Returns/creates the ServiceReader for the current ServiceNode

        Returns:
            ServiceReader: ServiceReader object for current ServiceNode
        """
        if not self._reader:
            self._reader = ServiceReader(self)
        return self._reader

    def print_resource_node(self, resource_node_name:str) -> None:
        # TODO: check if resource node exists
        operations_panel = self._get_operation_details_panel(resource_node_name)
        if operations_panel:
            console.print(operations_panel)
        
    def _get_operation_details_panel(self, resource_node_name:str) -> Panel:
        resource_node =  self.get_resource_node_by_name(resource_node_name)
        if not resource_node:
            return False
        panels_for_operations = [
            resource_node._rich_operation_details_panel(operation_name)
            for operation_name in resource_node.operation_names
        ]

        operations_group = Group(
            *panels_for_operations
        )
        return Panel(operations_group, title=f'[bold]Resource Node: [blue]{resource_node.name} ', title_align='left')
        
    def get_resource_node_by_name(self, resource_node_name:str) -> ResourceNode:
        """Searches the current ServiceNode for the given `resource_node_name`,
        and returns it.

        Args:
            resource_node_name (str): Name of the ResourceNode

        Returns:
            ResourceNode: The ResourceNode object, `False` if not found.
        """
        if not resource_node_name:
            return False
        
        for r_node in self.get_resource_nodes():
            # FIXME: laterr
            # if r_node.name == resource_node_name:
            if icompare_two_camel_case_words(r_node.name, resource_node_name):
                return r_node
        return False
    
    def json(self) -> Dict:
        return {
            "service_name": self.name,
            # TODO:
        }

    # TODO: memoization
    def find_resource_node_by_operation_name(self, operation_name:str) -> ResourceNode:
        """Traverses the ResourceNodes of the current ServiceNode and tries to find
        the ResourceNode that has the `operation_name` in it.

        Args:
            operation_name (str): Name of the operation

        Returns:
            ResourceNode: ResourceNode object that has the `operation_name`, or None.
        """
        for r_node in self.get_resource_nodes():
            if operation_name in r_node.operation_names:
                return r_node
        return None

    def get_relation_map(self) -> RelationMap:
        """Gets the relation map object.

        Returns:
            RelationMap: RelationMap object for the current ServiceNode
        """
        if self._relation_map is not None:
            return self._relation_map
        self._relation_map = RelationMap(self)
        return self._relation_map

    def get_resource_nodes(self) -> List[ResourceNode]:
        """Gets the available `ResourceNode`s of the current ServiceNode

        Returns:
            List[ResourceNode]: List of `ResourceNode`s available in the ServiceNode
        """
        if self.resource_nodes is None:
            self.resource_nodes = self._generate_resource_nodes()
        return self.resource_nodes

    def create_resource_node(self, **kwargs: Dict) -> ResourceNode:
        """Creates the ResourceNode object with the given `kwargs`.
        Uses the `ResourceNodeRegistry` to find the custom subclasses of the
        `ResourceNode` class, else defaults to use the ResourceNode class

        Raises: # TODO: fix this, return false instead
            Exception: _description_

        Returns:
            ResourceNode: _description_
        """
        service_name = self.name
        resource_node_name = kwargs.get('name', False)
        _ResourceNodeClass = ResourceNode
        if not resource_node_name:
            logger.debug(f"A resource name must be provided on creation.", extra=kwargs)
            raise Exception('ResourceNode name must be provided.')
        _custom_cls_for_resource_node = _resource_node_registry.find_custom_class_for_resource_node(service_name, resource_node_name)
        if _custom_cls_for_resource_node:
            logger.debug(f"ResourceNodeRegistry: [bold][green]{service_name}[/].[blue]{resource_node_name}[/][/] has extended with: {_custom_cls_for_resource_node}")
            _ResourceNodeClass = _custom_cls_for_resource_node
        return _ResourceNodeClass(**kwargs)
    
    def _generate_resource_nodes(self) -> List[ResourceNode]:
        """Parses `botocore` client for the AWS service and generates ResourceNodes.
        Generated `ResourceNode`s can be subclasses of `ResourceNode`.

        Returns:
            List[ResourceNode]: List of generated `ResourceNode`s
        """
        generated_resouce_nodes = []
        op_name_to_tokens_map = self.get_read_operation_name_to_tokens_map()
        resource_name_to_operations_map = {}
        for operation_name, op_name_tokens in op_name_to_tokens_map.items():
            verb, *resource_name_list = op_name_tokens
            resource_node_name = ''.join(resource_name_list)
            
            found_key = ifind_key_in_dict_keys(resource_node_name, resource_name_to_operations_map.keys())
            if found_key:
                # decide if we want to keep the plural or singular name
                
                sp_comparison = compare_nouns(resource_node_name, found_key)
                if sp_comparison == 's:p':
                    # swap the found_key with resource_node_name as dict key
                    resource_name_to_operations_map[resource_node_name] = [operation_name] + resource_name_to_operations_map[found_key]
                    resource_name_to_operations_map.pop(found_key)
                else:
                    resource_name_to_operations_map[found_key].append(operation_name)                
            else:
                resource_name_to_operations_map[resource_node_name] = [
                    operation_name]

        # createt the resource nodes
        for resource_node_name, operations_list in resource_name_to_operations_map.items():
            created_resource_node = self.create_resource_node(
                service_node=self, 
                name=resource_node_name, 
                operation_names=operations_list
            )
            if created_resource_node:
                generated_resouce_nodes.append(created_resource_node)

        if len(generated_resouce_nodes) <= 1:
            return generated_resouce_nodes

        # TODO: decide if this should be kept
        # concat same names
        combined_resource_nodes = []
        _used_for_combining_resource_nodes = []

        for i, gen_resource_node in enumerate(generated_resouce_nodes[:-1]):
            if gen_resource_node in _used_for_combining_resource_nodes:
                continue
            for other_resource_node in generated_resouce_nodes[i+1:]:
                # FIXME: IMPORTANT
                if compare_two_camel_case_words(gen_resource_node.name, other_resource_node.name):
                    # select the shortest name, meaning singular
                    singular_named_resource_node = gen_resource_node
                    plural_named_resource_node = other_resource_node
                    if len(singular_named_resource_node.name) > len(plural_named_resource_node.name):
                        singular_named_resource_node, plural_named_resource_node = plural_named_resource_node, singular_named_resource_node
                    _used_for_combining_resource_nodes.append(
                        plural_named_resource_node)
                    singular_named_resource_node.operation_names.extend(
                        plural_named_resource_node.operation_names)

            if gen_resource_node not in _used_for_combining_resource_nodes:
                combined_resource_nodes.append(gen_resource_node)

        return combined_resource_nodes

    def get_read_operation_name_to_tokens_map(self) -> Dict:
        """Generate `operation name to word tokens` map for the 
        available read operations in the ServiceNode. 
        
        Caches the output in the class attr. `_read_operation_name_to_tokens_map`.

        Returns:
            Dict: dictionary 
        """
        if self._read_operation_name_to_tokens_map is not None:
            return self._read_operation_name_to_tokens_map
        read_operation_names = self.get_read_operation_names()
        self._read_operation_name_to_tokens_map = self._generate_operation_name_to_tokens_map(
            read_operation_names)
        return self._read_operation_name_to_tokens_map

    def _generate_operation_name_to_tokens_map(self, operation_names:List[str]=None) -> Dict:
        if operation_names is None:
            operation_names = self.get_operation_names()

        op_name_to_tokens = {
            op_name: camel_case_split(op_name)
            for op_name in operation_names
        }
        return op_name_to_tokens

    def get_read_operation_names(self) -> List[str]:
        operation_names = self.get_operation_names()
        read_only_operation_names = [
            op_name
            for op_name in operation_names
            if any([op_name.startswith(read_verb) for read_verb in READ_ONLY_VERBS])
        ]
        return read_only_operation_names

    def get_service_model(self) -> ServiceModel:
        service_model = self.client._service_model
        return service_model

    def get_event_system(self) -> EventAliaser:
        event_system = self.client.meta.events
        return event_system

    def get_operation_names(self) -> List[str]:
        """Returns the `operation_names` defined in the boto client._service_model"""
        service_model = self.get_service_model()
        operation_names = service_model.operation_names
        return operation_names



try:
    from ..nodes import ResourceNode
    from ..logs import get_logger
except ImportError:
    from nodes import ResourceNode
    from logs import get_logger
logger = get_logger(__name__)


class SecurityGroups(ResourceNode, service_name="ec2", name="SecurityGroups"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


"""
1. check for already existing operation data
2. get relations of the operation
3. find the dependent operations to call
4. read the target dependent operation
5. generate JMESpath selectors with the relations
6. search with JMESpath to get raw_parameters
7. generate real api parameters from raw_parameters
8. read the current operation
"""



try:
    from .utils import icompare_two_camel_case_words, str_relations, ifind_similar_names_in_list
    from .botocore_utils import READ_ONLY_VERBS
    from .relations import FindRelationResultTypes, SUCCESSFUL_FIND_RELATION_RESULT_TYPES
    from .logs import get_logger, get_rich_console
    from .errors import Error
except ImportError:
    from utils import icompare_two_camel_case_words, str_relations, ifind_similar_names_in_list
    from botocore_utils import READ_ONLY_VERBS
    from relations import FindRelationResultTypes, SUCCESSFUL_FIND_RELATION_RESULT_TYPES
    from logs import get_logger, get_rich_console
    from errors import Error
import fnmatch
from collections.abc import Iterable
import jmespath
from rich.progress import Progress, track, BarColumn, TextColumn, TaskProgressColumn, TimeRemainingColumn
from rich.table import Column
logger = get_logger(__name__)
console = get_rich_console()
from typing import List, Set, Dict, Tuple, Optional, Union
class ServiceReader:
    def __init__(self, service_node):
        self.service_node = service_node
        self.response_data = {}

    def add_response_data(self, operation_name, response):
        if operation_name not in self.response_data:
            self.response_data[operation_name] = [response]
        else:
            self.response_data[operation_name].append(response)
            

    def verify_all_required_parameters_in_selected_relations(self, required_parameters, selected_relations):
        if not selected_relations or type(selected_relations) not in (list,tuple):
            return False
        parameter_to_relation_map = {required_parameter_name: None for required_parameter_name in required_parameters}
        for required_parameter_name in parameter_to_relation_map.keys():
            is_parameter_in_relation_map = any([
                icompare_two_camel_case_words(required_parameter_name, relation_dict.get('search_shape_name')) 
                 for relation_dict in selected_relations
            ])
            parameter_to_relation_map[required_parameter_name]=is_parameter_in_relation_map
        is_all_parameters_found_in_relations = all(parameter_to_relation_map.values())
        return is_all_parameters_found_in_relations
    
    
    def _call_operation(self, operation_name, api_parameters):
        # TODO: follow next tokens
        client = self.service_node.client
        from botocore.exceptions import ClientError
        try:
            # print(colored(f"{operation_name=} {api_parameters=}",'green'))
            logger.debug(f"Calling operation: [bold blue]{operation_name}[/] with api parameters: {api_parameters}")
            response = client._make_api_call(operation_name, api_parameters)
            # response = None
            response
        except ClientError as e:
            # print(colored(str(e),'red'))
            logger.debug(f"FAILED: CALLING OPERATION. {operation_name} with api parameters: {api_parameters}. Exception: {str(e)} ")
            return False
        
        resource_node = self.service_node.find_resource_node_by_operation_name(operation_name)
        if not resource_node:
            return False
        resource_node_name = resource_node.name
        self.add_to_node_data(resource_node_name, operation_name, response)
        response.pop('ResponseMetadata') # FIXME: add back
        response['__args__'] = api_parameters
        return response
    
    
        # if operation_data:
        #     new_api_params = jmespath.search(flattened_jmespath_nested_selector, operation_data)
        # operation_data
        # for 
        # # jmespath.search(f"[*].{target_path}[][]", operation_data)
        
        # target_paths = target_path.split(',')
        # all_found = []
        # for target_path in target_paths:
        #     # found =  
        #     jmespath.search(f"{_for_all_the_responses}{target_path}{_flatten_two_times}", response)
        #     if found:
        #         all_found.extend(found)
        # return all_found

    def search_operation_data(self, resource_node_name, operation_name):
        resource_node_exists = self.response_data.get(resource_node_name, False) != False
        if not resource_node_exists:
            return False
        result = self.response_data[resource_node_name].get(operation_name, False)
        return result
    
    def search_resource_node_data(self, resource_node_name):
        resource_node_data = self.response_data.get(resource_node_name, False) 
        return resource_node_data 
   
    
    def add_to_node_data(self, resource_node_name, operation_name, response):
        resource_node_exists = self.response_data.get(resource_node_name, False) != False
        
        if not resource_node_exists:
            self.response_data[resource_node_name] = {}
                
        if operation_name not in self.response_data[resource_node_name]:
            self.response_data[resource_node_name][operation_name] = [response]
        else:
            self.response_data[resource_node_name][operation_name].append(response)


    def read_operation(self, resource_node_name:str, operation_name:str, match_patterns=None, refresh:bool=False) -> Tuple[Union[List, bool], Union[Error, None]]:
        # if it has been read called already, return it

        resource_node = self.service_node.get_resource_node_by_name(resource_node_name)

        if refresh == True:
            pass # TODO: delete from reader
        already_existing_data = self.search_operation_data(resource_node.name, operation_name)
        
        if already_existing_data!=False and refresh == False:
            return already_existing_data
        
        
        if not resource_node:
            logger.debug(f"RESOURCE NODE NOT FOUND. While reading the {resource_node_name}.{operation_name}, Resource Node {resource_node_name} could not be loaded.")
            return False

        # add service_name and resource_node_name to relation dict
        relations_of_operation, relations_error = resource_node.get_operations_relations(operation_name)
        success_finding_relations = relations_error is None
        
        
        def api_parameters_match_pattern(api_parameters, patterns):
            if not patterns:
                return api_parameters
            matched_parameters = []
            
            for api_param in api_parameters:
                has_api_param_added = False
                for api_param_value in api_param.values():
                    for pattern in patterns:
                        if not has_api_param_added and fnmatch.fnmatch(api_param_value, pattern) \
                                and api_param not in matched_parameters:
                            matched_parameters.append(api_param)
                            has_api_param_added = True
            return matched_parameters
        
            
        if not success_finding_relations:
            logger.debug(f"FAILED FINDING RELATIONS for OPERATION: [bold]{resource_node_name}.{operation_name}[/]. Failed to find relations: {relations_error}")
            return False, relations_error
        
        if relations_of_operation == True:
            # no required parameters
            generated_api_parameters, generation_error = resource_node.generate_api_parameters_from_operation_data(operation_name, [], {})
            
            if generation_error is not None:
                return False, generation_error
            if isinstance(generated_api_parameters, Iterable):
                # TODO: filter generated_api_parameters if a pattern is given.
                
                pattern_matched_api_parameters = api_parameters_match_pattern(generated_api_parameters, match_patterns)
                for api_parameters in pattern_matched_api_parameters:
                    self._call_operation(operation_name, api_parameters)
            return self.search_operation_data(resource_node_name, operation_name)

        all_related_operations_data = {}
        for rel in relations_of_operation:
            rel_operation_data = self.read_operation(rel.get('resource_node_name'), rel.get('operation_name'), refresh=refresh)
            if not rel_operation_data:
                logger.debug(f"[red]FAILED TO FIND RELATED RESOURCES[/] from [bold]{resource_node_name}[/], while reading:  [bold]{rel.get('service_name')}.{rel.get('resource_node_name')}.{rel.get('operation_name')}[/] operation.")
                return False
          
            # gather all their related data, put it under a dict 
            all_related_operations_data.update({
                rel.get('operation_name'): rel_operation_data
            })

        # send the operations_data to resource_node to create valid_api_parameters
        generated_api_parameters, generation_error = resource_node.generate_api_parameters_from_operation_data(operation_name, relations_of_operation, all_related_operations_data)

        if generated_api_parameters == []:
            logger.debug(f"FAILED TO AUTO-GENERATE API PARAMETERS. Related Resources couldn't found.")
        elif isinstance(generated_api_parameters, Iterable):
            # FIXME # for api_parameters in track(generated_api_parameters, description=f"Calling [green]{operation_name}[/] for [bold]{len(generated_api_parameters)}[/] resources...",transient=True, console=console):


            # TODO: filter generated_api_parameters if a pattern is given.
            pattern_matched_api_parameters = api_parameters_match_pattern(generated_api_parameters, match_patterns)
            for api_parameters in pattern_matched_api_parameters:
                self._call_operation(operation_name, api_parameters)
        else:
            logger.debug(f"COULDN'T GENERATE API PARAMETERS. {resource_node_name}.{operation_name}. Generated Parameters: {generated_api_parameters}. Data: {all_related_operations_data}")
            
        return self.search_operation_data(resource_node_name, operation_name)


            
    def read_resource_node(self, resource_node_name, match_patterns=None):
        resource_node = self.service_node.get_resource_node_by_name(resource_node_name)
        if not resource_node:
            return False
        # print('\t', colored(resource_node_name,'yellow'))
        # read all operations
        
        for operation_name in resource_node.operation_names:
            # if operation_name != 'GetPasswordData':
            #     continue
            req_params = resource_node.get_required_parameter_names_from_operation_name(operation_name)
            # if req_params:
            req_params
            self.read_operation(resource_node_name, operation_name, match_patterns)
            # print('--------------------------------')
            # print()
          
        # return self.response_data.get(resource_node.name)
        return self.search_resource_node_data(resource_node.name)
            # if output == True:
            #     print(' >>>>>>>>>>>>>>>> NO PARAMETERS <<<<<<<<<<<<<<<<<<<<<<')


 



try:
    from .custom_nodes import *
    from .nodes import *
    from .logs import get_logger
    from .factories import Boto3SessionSingleton
    from .utils import get_all_available_services, _create_boto_session
    from .registries import AppRegistry, app_registry
except ImportError:
    from custom_nodes import *
    from nodes import *
    from logs import get_logger, get_rich_console
    from factories import Boto3SessionSingleton
    from utils import get_all_available_services, _create_boto_session
    from registries import AppRegistry, app_registry

    from custom_nodes import *


import boto3
import sys

logger = get_logger(__name__)
console = get_rich_console()

# print(app_registry.app_configs)
# print(app_registry.app_configs)
# x = AppConfig.create('balcony_app.myservice')
# print(x)
# x




"""
1. check for already existing operation data
2. get relations of the operation
3. find the dependent operations to call
4. read the target dependent operation
5. generate JMESpath selectors with the relations
6. search with JMESpath to get raw_parameters
7. generate real api parameters from raw_parameters
8. read the current operation
"""
if __name__ == '__main__':
    # session = boto3.session.Session()
    session = Boto3SessionSingleton().get_session()
    all_service_names = get_all_available_services(session)
    success_count, failure_count, non_req_count = 0,0, 0
    
    service_node = ServiceNode('ec2', session)
    # service_node = ServiceNode('accessanalyzer', session)
    # service_reader = ServiceReader(service_node)

    # aa = service_reader.read_operation('PasswordData', 'GetPasswordData')
    # aa = service_reader.read_operation('AccessAnalyzer', 'GetAccessAnalyzer')
    # aa = service_reader.read_operation('NetworkInsightsAccessScopeAnalyses', 'DescribeNetworkInsightsAccessScopeAnalyses')
    # aa = service_reader.read_operation('ChangeSet', 'ListChangeSets')
    
    # aa

    # aa = service_reader.read_operation('Instances', 'DescribeInstances')
    # aa = service_reader.read_operation('SecurityGroups', 'DescribeSecurityGroups')
    # if aa:
    #     aa
    # for service_name in all_service_names:
        # continue
    # for service_name in []:
    for service_name in ['s3']:
        service_node = ServiceNode(service_name, session)
        service_node
        
   

        console.print('-'*50)
        console.print('#'*10, colored(service_name, 'yellow'), '#'*10)
        # resource_nodes = service_node.get_resource_nodes()
        service_reader = service_node.get_service_reader()


        read_operation_names = service_node.get_read_operation_names()
        for r_ops_name in read_operation_names:
            verb, *resource_name_tokens = camel_case_split(r_ops_name)
            resource_name = ''.join(resource_name_tokens)
            a = service_reader.read_operation(resource_name, r_ops_name)
    # #         # x = relation_map.find_best_relations_for_operations_parameters(r_ops_name)
            console.print(a)
            console.print('-'*40)
    #         if x:
    #             if len(x) == 1:
    #                 non_req_count+= 1
    #             success_count += 1 
    #         else:
    #             failure_count += 1
            
            
    #         # get_required_parameter_names_from_operation_name()
    #         print_flag = False
    #         if type(x)==dict:
    #             for _item in x.values():
    #                 if type(_item) in [dict, list]:
    #                     print_flag = True
    #                     break
    #         if print_flag:
    #             print(json.dumps(x, indent=2, default=str))
    #         if x:
    #             print( json.dumps(x, indent=2, default=str))
    #         else:
    #             print(colored('\t'+r_ops_name+'   FALSE', 'red'))
    #     print('SUCCESS', success_count)
    #     print('FAILURE', failure_count)
    #     print('NON_REQUIRED_PARAMS', non_req_count)
    #     print('TOTAL', success_count+failure_count)


from pprint import pprint
from typing import Union
import sys
from importlib import import_module
from typer.models import TyperInfo
from typer.main import Typer
try:
    from .registries import app_registry
except ImportError:
    from registries import app_registry

APPS_MODULE_NAME = "app"

class BaseBalconyApp:
    author = None
    app_name = None
    description = None
    tags = ()
    
    def __init_subclass__(cls, author=None, app_name=None, description=None, tags=(), **kwargs):
        super().__init_subclass__(**kwargs)
        cls.author = author
        cls.app_name = app_name
        cls.description =description
        cls.tags = tags
        app_registry.register_app_class(cls, author, app_name, description, tags, **kwargs)

        

            
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        print('BaseBalconyApp initialized.')
        # self.name = name
        

    def get_data(self, *args, **kwargs) -> dict:
        """"""
        raise NotImplementedError("Please Implement this method")
    
    def get_cli_app(self, *args, **kwargs) -> Typer:
        """"""
        # TODO: create default cli app on template repository
        raise NotImplementedError("Please Implement this method")
        
# class BaseBalconyAppResource:
#     def __init__(self, name, *args, **kwargs) -> None:
#         self.name = name

"""
balcony-app/
    


"""


import boto3
try:
    from .utils import get_all_available_services, _create_boto_session
    from .nodes import ServiceNode
except ImportError:
    from utils import get_all_available_services, _create_boto_session
    from nodes import ServiceNode
    
class Boto3SessionSingleton(object):
    _instance = None
    _session = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Boto3SessionSingleton, cls).__new__(cls)
            # Put any initialization here.
            # cls._session = _create_boto_session()

        return cls._instance


    def __init__(self):
        if not self._session:
            self._session = _create_boto_session()
    
    def get_session(self):
        return self._session

class ServiceNodeFactory:
    def __init__(self, boto3_session):
        self.boto3_session = boto3_session
        self._service_nodes_map = {}

    def _create_service_node(self, service_name):
        service_node = ServiceNode(service_name, self.boto3_session)
        self._service_nodes_map[service_name] = service_node

    def get_service_node(self, service_name):
        if not service_name in self._service_nodes_map:
            self._create_service_node(service_name)            
        return self._service_nodes_map.get(service_name)
    





class BalconyAWS:
    def __init__(self, boto3_session=None) -> None:
        if boto3_session is None:
            self.boto3_session = Boto3SessionSingleton().get_session()
        else:
            self.boto3_session = boto3_session
        self.service_node_factory = ServiceNodeFactory(self.boto3_session)
        
    def get_available_service_node_names(self):
        return get_all_available_services(self.boto3_session)
    
    
    def get_service(self, service_name:str) -> ServiceNode:
        service_node = self.service_node_factory.get_service_node(service_name)
        return service_node
    




try:
    from ..nodes import ResourceNode
    from ..logs import get_logger
    from ..errors import Error
except ImportError:
    from nodes import ResourceNode
    from logs import get_logger
    from errors import Error

logger = get_logger(__name__)
from typing import List, Set, Dict, Tuple, Optional, Union



class Policy(ResourceNode, service_name="iam", name="Policy"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "service_name": "iam",
            "resource_node_name": "Policy",
            "search_shape_name": "PolicyArn",
            "target_shape_name": "Arn",
            "target_shape_type": "string",
            "operation_name": "ListPolicies",
            "target_path": "Policies[*].Arn"
        },
        {
            "service_name": "iam",
            "resource_node_name": "Policy",
            "search_shape_name": "VersionId",
            "target_shape_name": "DefaultVersionId",
            "target_shape_type": "string",
            "operation_name": "ListPolicies",
            "target_path": "Policies[*].DefaultVersionId"
        }]
        
    
    def generate_api_parameters_from_operation_data(self, operation_name:str, 
                                                    relations_of_operation:List[Dict], 
                                                    related_operations_data: Union[List, Dict]) -> Tuple[Union[List, bool], Union[Error, None]]:
        generated_api_parameters, err = super().generate_api_parameters_from_operation_data(operation_name, relations_of_operation, related_operations_data)

        if err is not None:
            return generated_api_parameters, err
        
        if operation_name == "ListPolicies":
            api_parameters = []
            for api_param in generated_api_parameters:
                api_param.update({'Scope':'Local'})
                api_parameters.append(api_param)
            return api_parameters, None
            
        return generated_api_parameters, err
            

class User(ResourceNode, service_name="iam", name="User"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "service_name": "iam",
            "resource_node_name": "User",
            "search_shape_name": "UserName",
            "target_shape_name": "UserName",
            "target_shape_type": "string",
            "operation_name": "ListUsers",
            "target_path": "Users[*].UserName"
        }]

class Role(ResourceNode, service_name="iam", name="Role"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "service_name": "iam",
            "resource_node_name": "Role",
            "search_shape_name": "RoleName",
            "target_shape_name": "RoleName",
            "target_shape_type": "string",
            "operation_name": "ListRoles",
            "target_path": "Roles[*].RoleName"
        }]
    
class AccessKeys(ResourceNode, service_name="iam", name="AccessKeys"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "service_name": "iam",
            "resource_node_name": "AccessKeys",
            "search_shape_name": "AccessKeyId",
            "target_shape_name": "AccessKeyId",
            "target_shape_type": "string",
            "operation_name": "ListAccessKeys",
            "target_path": "AccessKeyMetadata[*].AccessKeyId"
        }]
        


import typer
try:
    from .utils import *
    from .logs import get_logger, get_rich_console, set_log_level_at_runtime
    from .nodes import ServiceNode, OperationType
    from .reader import ServiceReader
    from .custom_nodes import *
    from .registries import app_registry
    from .factories import Boto3SessionSingleton, BalconyAWS
    from .settings import INSTALLED_BALCONY_APPS
except ImportError:
    from utils import *
    from logs import get_logger, get_rich_console, set_log_level_at_runtime
    from nodes import ServiceNode, OperationType
    from reader import ServiceReader
    from custom_nodes import *
    from factories import Boto3SessionSingleton, BalconyAWS
    from registries import app_registry
    from settings import INSTALLED_BALCONY_APPS

from typing import Optional
import boto3
from rich.columns import Columns
from rich.panel import Panel
from rich.console import Console
from rich.layout import Layout
from rich.pretty import Pretty
import logging

console = get_rich_console()
err_console = Console(stderr=True)
logger = get_logger(__name__)
session = Boto3SessionSingleton().get_session()
balcony_aws = BalconyAWS(session)
app = typer.Typer(no_args_is_help=True)
# aws_app = typer.Typer() # no_args_is_help=True
awsx_app = typer.Typer(no_args_is_help=True) # no_args_is_help=True
# app.add_typer(aws_app, name="aws")
app.add_typer(awsx_app, name="apps")
app_registry.import_balcony_apps(INSTALLED_BALCONY_APPS)


app_objects = []
registered_apps = app_registry.get_registered_apps()
for author, app_dict_list in registered_apps.items():
    for app_dict in app_dict_list:
        app_cls = app_dict.get('cls')
        app_name = app_dict.get('app_name')
        app_obj = app_cls()
        app_objects.append(app_obj)

        try:
            _typer_app = app_obj.get_cli_app()
            if _typer_app:
                awsx_app.add_typer(_typer_app, name=app_name)
        except NotImplementedError:
            print(app_name, 'has not implemented get_cli_app()') 
            

def _get_available_service_node_names():
    return get_all_available_services(session)

def generate_service_node_completion_items():
    service_names = _get_available_service_node_names(session)
    return service_names

def _complete_service_name(incomplete: str):
    service_names = _get_available_service_node_names()
    if not incomplete:
        for name in service_names:
            yield name
    else:
        for name in service_names:
            if name.startswith(incomplete):
                yield name
                
                
def _complete_resource_node_name(ctx: typer.Context, incomplete: str):
    service = ctx.params.get("service", False)
    if not service:
        return []
    service_node = balcony_aws.get_service(service)
    if not service_node:
        return []
    resource_nodes = service_node.get_resource_nodes()
    if not resource_nodes:
        return []
    
    resource_node_names = []
    
    for _rn in resource_nodes:
        _rn_name = _rn.name
        if incomplete:
            if _rn_name.lower().startswith(incomplete.lower()):
                resource_node_names.append(_rn_name)
        else:
            resource_node_names.append(_rn_name)
    return resource_node_names

def _complete_operation_type(ctx: typer.Context):
    # service, resource_node = 'ec2', 'BundleTasks'
    service = ctx.params.get("service", False)
    resource_node = ctx.params.get("resource_node", False)
    print(service, resource_node)
    if not service or not resource_node:
        return []
    service_node = balcony_aws.get_service(service)
    if not service_node:
        return []
    resource_node_obj = service_node.get_resource_node_by_name(resource_node)
    if not resource_node_obj:
        return []

    operation_types_and_names = resource_node_obj.get_operation_types_and_names()
    print(f"{operation_types_and_names=}")
    return list(operation_types_and_names.keys())
        
        
# @aws_app.command('ls')
def _list_service_or_resource(
        service: Optional[str] = typer.Argument(None, show_default=False,help='The AWS service name', autocompletion=_complete_service_name),
        resource_node: Optional[str] = typer.Argument(None, show_default=False, help='The AWS Resource Node', autocompletion=_complete_resource_node_name),
    ):

    available_service_names = _get_available_service_node_names()
    if not service and not resource_node:
        # nothing is given
        console.print(Columns(available_service_names, equal=True, expand=True))
        return available_service_names
    
    if service and not resource_node:
        # we only have service name
        
        if service not in available_service_names:
            similar_service_names = ifind_similar_names_in_list(service, available_service_names)
            if similar_service_names:
                console.print(f"[yellow]Found Similar Service Names:")
                console.print(Columns(similar_service_names, equal=True, expand=True))
                return
            else:
                raise typer.Exit(f"Invalid service name: {service}. Please pick a proper one.")
            
        service_node = balcony_aws.get_service(service)
        resource_nodes = service_node.get_resource_nodes()
        
        resource_node_names = []
        for _rn in resource_nodes:
            _rn_name = _rn.name
            if _rn.get_all_required_parameter_names():
                _rn_name = f"[bold]{_rn_name}[/]"
            resource_node_names.append(_rn_name)
        
        resource_node_name_as_columns =  Columns(resource_node_names, equal=True, expand=True)
        console.print(resource_node_name_as_columns)
        return resource_node_names
        
    elif service and resource_node:
        # we got both options filled
        service_node = balcony_aws.get_service(service) 
        resource_nodes = service_node.get_resource_nodes()
        resource_node_names = [ _rn.name for _rn in resource_nodes]
        resource_node_obj = service_node.get_resource_node_by_name(resource_node)
        if not resource_node_obj:
            
            similar_resource_names = ifind_similar_names_in_list(resource_node, resource_node_names)
            if similar_resource_names:
                console.print(f"[yellow]Found Similar Resource Node Names:")
                console.print(Columns(similar_resource_names, equal=True, expand=True))
                return 
            else:
                raise typer.Exit(f"Invalid Resource Node name: [bold]{resource_node}[/] for Service: [bold]{service}[/].")
        operations_panel = service_node._get_operation_details_panel(resource_node_obj.name)
        console.print(operations_panel)
        return    

def aa():
    # for sname in balcony_aws.get_available_service_node_names():
    a = _complete_operation_type(None, None)
    print(a)
    
    # service_node = balcony_aws.get_service('iam')
    # for rn in service_node.get_resource_nodes():
    #     for operation in rn.get_operation_names():
    #         print(operation)
    
@app.command('aws')
def aws_main_command(
        service: Optional[str] = typer.Argument(None, show_default=False,help='Name of the AWS Service', autocompletion=_complete_service_name),
        resource_node: Optional[str] = typer.Argument(None, show_default=False, help='Name of the AWS Resource Node', autocompletion=_complete_resource_node_name),
        operation: Optional[str] = typer.Argument(None, show_default=False,case_sensitive=False , show_choices=False,help='Select a specific operation type.', autocompletion=_complete_operation_type),
        patterns: Optional[List[str]] = typer.Option(None, "--pattern", '-p', show_default=False, help='UNIX pattern matching for generated parameters. Should be quoted. e.g. (-p "*prod-*")'),
        list_contents: bool = typer.Option(False, "--list", '-l', help='Print the details of Service or Resource. Does not make requests.'),
        debug: bool = typer.Option(False, "--debug", '-d', help='Enable debug messages.'),
        paginate: bool = typer.Option(False, "--screen", '-s', help='Open the data on a separate paginator on shell.'),
    ):
    if debug:
        set_log_level_at_runtime(logging.DEBUG)
    
    if list_contents:
        _list_service_or_resource(service, resource_node)
        return
    # aa()
    # return
    # available_service_names = _get_available_service_node_names()
    if not service and not resource_node:
        available_services = _list_service_or_resource(service, resource_node)
        console.print(Panel(f"[bold]Please pick one of the AWS Services", title="[red][bold]ERROR"))
        return {'services':available_services}
    
    if service and not resource_node:
        available_resources = _list_service_or_resource(service, resource_node)
        console.print(Panel(f"[bold]Please pick one of the Resource Nodes from [green]{service}[/] Service", title="[red][bold]ERROR"))
        return {'service':service, 'resources': available_resources}
    elif service and resource_node:
        service_node = balcony_aws.get_service(service)
        service_reader = service_node.get_service_reader()
        
        is_operation_selected = operation is not None
        read_data = None
        if not is_operation_selected:
            read_data = service_reader.read_resource_node(resource_node, match_patterns=patterns)
          
        else: # Operation is selected
            
            resource_node_obj = service_node.get_resource_node_by_name(resource_node)
            types_to_op_names = resource_node_obj.get_operation_types_and_names()
            supported_operation_types = list(types_to_op_names.keys())
            operation_name = types_to_op_names.get(operation.value, False)
            if not operation_name:
                console.print(f"[red bold]Given {operation.value} is not supported by {resource_node}. Try: {supported_operation_types}")
                return False
            console.print(patterns)
            read_data = service_reader.read_operation(resource_node, operation_name, match_patterns=patterns)

        if paginate:
            with console.pager(styles=True):
                console.print_json(data=read_data, default=str)
        else:    
            console.print_json(data=read_data, default=str)
        return read_data

    
def run_app():
    app(prog_name="balcony")
    

    
if __name__ == "__main__":
    run_app()
    


import os
from pathlib import Path

HOME_DIR = os.path.expanduser('~')
BALCONY_CONFIG_DIR = os.getenv('BALCONY_CONFIG_DIR', os.path.join(HOME_DIR, '.balcony'))
BALCONY_RELATIONS_DIR = os.getenv('BALCONY_RELATIONS_DIR', os.path.join(BALCONY_CONFIG_DIR, 'relations'))

# create the relations directory if not exists
Path(BALCONY_RELATIONS_DIR).mkdir(parents=True, exist_ok=True)


LOG_LEVEL = 'INFO'

INSTALLED_BALCONY_APPS = (
    'balconyapp',
)


try:
    from ..nodes import ResourceNode
    from ..logs import get_logger
except ImportError:
    from nodes import ResourceNode
    from logs import get_logger
logger = get_logger(__name__)



class Cluster(ResourceNode, service_name="ecs", name="Clusters"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def generate_jmespath_selector_from_relations(self, operation_name, relation_list):
        r = super().generate_jmespath_selector_from_relations(operation_name, relation_list)
        if operation_name == 'ListClusters':
            return '[*].{cluster: clusterArns}[][]'
        return r
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "search_shape_name": "cluster",
            "target_shape_name": "cluster",
            "target_shape_type": "string",
            "operation_name": "ListClusters",
            "target_path": "clusterArns"
        }]
        
        
class Services(ResourceNode, service_name="ecs", name="Services"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_required_parameter_names_from_operation_name(self, operation_name):
        r = super().get_required_parameter_names_from_operation_name(operation_name)
        if operation_name == 'ListServices':
            return ['cluster']
        return r

    def get_operations_relations(self, operation_name):
        r = super().get_operations_relations(operation_name)
        if operation_name == 'DescribeServices':
            return [{
                "search_shape_name": "services",
                "target_shape_name": "serviceArns",
                "target_shape_type": "string",
                "operation_name": "ListServices",
                "target_path": "serviceArns"
            },
            {
                "search_shape_name": "cluster",
                "target_shape_name": "cluster",
                "target_shape_type": "string",
                "operation_name": "ListServices",
                "target_path": "__args__.cluster"
            }
            ], None
        return r
    
    # def define_extra_relations(self):
    #     r= super().define_extra_relations()
   


"""
When a Python module is imported as `from <module> import *`,
modules `__init__.py` file is imported.  

"""

from .ec2 import *
from .codebuild import *
from .ecs import *
from .lambda_functions import *
from .iam import *
from .ssm import *
from .s3 import *


"""
Any resource that app system is going to use must be exported here.
Otherwise different import statements would create new global static objects
for AppRegistry etc.
"""
try:
    from .registries import app_registry, AppRegistry
    from .app import BaseBalconyApp
    from .factories import BalconyAWS
    from .logs import get_rich_console, get_logger
except ImportError:
    from registries import app_registry, AppRegistry
    from app import BaseBalconyApp
    from factories import BalconyAWS
    from logs import get_rich_console, get_logger



try:                                                                                                                                                                                       
    from ..nodes import ResourceNode                                                                                                                                                       
    from ..logs import get_logger                                                                                                                                                          
except ImportError:                                                                                                                                                                        
    from nodes import ResourceNode                                                                                                                                                         
    from logs import get_logger                                                                                                                                                            
from typing import List, Set, Dict, Tuple, Optional, Union                                                                                                                                 
logger = get_logger(__name__)                                                                                                                                                              
                                                                                                                                                                                            
class BuildsForProject(ResourceNode, service_name='codebuild', name='BuildsForProject'):                                                                                                   
    def __init__(self, *args, **kwargs):                                                                                                                                                   
        super().__init__(*args, **kwargs)                                                                                                                                                  
                                                                                                                                                                                            
    def get_operations_relations(self, operation_name) -> Tuple[List[Dict], None]:                                                                                                               
        result, err = super().get_operations_relations(operation_name)                                                                                                                     
        # manipulate `result` or return your own value                                                                                                                                     
        return result, err


try:
    from ..nodes import ResourceNode
    from ..logs import get_logger
except ImportError:
    from nodes import ResourceNode
    from logs import get_logger
logger = get_logger(__name__)

class Function(ResourceNode, service_name="lambda", name="Function"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "service_name": "lambda",
            "resource_node_name": "Function",
            "operation_name": "ListFunctions",
            "search_shape_name": "FunctionName",
            "target_shape_name": "FunctionName",
            "target_shape_type": "string",
            "target_path": "Functions[*].FunctionName"
        }]


import logging
from rich.logging import RichHandler
from rich.console import Console

try:
    from .settings import LOG_LEVEL
except ImportError:
    from settings import LOG_LEVEL
    
_console = Console(color_system="auto", markup=True)
_balcony_loggers = []

def get_rich_console() -> Console:
    """Returns the global defined `rich.console.Console` object for common use. 

    Returns:
        rich.console.Console: rich console object
    """
    return _console

def supress_other_module_logs() -> None:
    """Sets Python modules log levels to `logging.CRITICAL` to supress them."""
    _supress_module_names = ('boto', 'urllib3', 's3transfer', 'boto3', 'botocore', 'nose')
    for _logger_name in logging.Logger.manager.loggerDict.keys():
        if _logger_name in _supress_module_names:
            logging.getLogger(_logger_name).setLevel(logging.CRITICAL)
            

def set_log_level_at_runtime(log_level):
    for _logger in _balcony_loggers:
        _logger.setLevel(log_level)
        # for handler in _logger.handlers:
        #     if isinstance(handler, type(logging.StreamHandler())):
        #         handler.setLevel(log_level)
        #         _logger.debug('Debug logging enabled')

def get_logger(name: str) -> logging.Logger:
    """Logger creation with RichHandler. 

    Args:
        name (str): Logger name. Usually given `__name__`.

    Returns:
        logging.Logger: Logger obj
    """
    supress_other_module_logs()
    logging.basicConfig(
        level=LOG_LEVEL,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True, markup=True, console=_console)]
    )
    # logging.FileHandler("debug.log", mode='w')
    
    _logger = logging.getLogger(name)
    _balcony_loggers.append(_logger)
    return _logger



try:
    from .botocore_utils import get_shape_name, flatten_shape_to_its_non_collection_shape_and_target_paths, IDENTIFIER_NAMES
    from .logs import get_logger
    from .utils import icompare_two_camel_case_words
    from .settings import BALCONY_RELATIONS_DIR
except ImportError:
    from botocore_utils import get_shape_name, flatten_shape_to_its_non_collection_shape_and_target_paths, IDENTIFIER_NAMES
    from logs import get_logger
    from utils import icompare_two_camel_case_words
    from settings import BALCONY_RELATIONS_DIR
from typing import List, Dict, Tuple, Union
import json
import os
from enum import Enum
logger = get_logger(__name__)

class FindRelationResultTypes(Enum):
    Success = 'Success'
    NoRequiredParameters = 'No required parameters'
    NoGeneratedParameters = 'NoParametersGenerated'
    CantDecideBetweenGeneratedParameters = 'CantDecideBetweenGeneratedParameters'
    SomeRelationsFoundButNotAll = 'SomeRelationsFoundButNotAll'
    NoRelations = 'NoRelations'
    RelationsFound = 'RelationsFound'
    InternalError = 'InternelError'

SUCCESSFUL_FIND_RELATION_RESULT_TYPES = (
    FindRelationResultTypes.Success,
    FindRelationResultTypes.NoRequiredParameters,
    FindRelationResultTypes.RelationsFound,
)
class RelationMap:
    def __init__(self, service_node):
        self.service_node = service_node
        self._relations_map: Dict = None

    def get_relations_map(self, refresh:bool=False) -> Dict:
        """Tries the fetch the cached RelationMap from file,
        or generates and saves it to file.

        Args:
            refresh (bool, optional): Disables cache. Defaults to False.

        Returns:
            Dict: _description_
        """
        if self._relations_map is not None and not refresh:
            return self._relations_map

        loaded_relations_map = self._load_relations_from_file()
        if loaded_relations_map and not refresh:
            self._relations_map = loaded_relations_map
            return self._relations_map

        generated_relations = self._generate_relation_map()
        self._relations_map = generated_relations
        self._save_relations_map_to_file()
        return self._relations_map

    # def check_parameter_name_inside_map(self, )
    def get_parameters_generated_relations(self, parameter_name:str, exclude_operation_name:str) -> List[Dict]:
        """_summary_

        Args:
            parameter_name (str): Parameter name to search for its relations
            exclude_operation_name (str): Exclude the given operation_name across found relation dicts

        Returns:
            List[Dict]: List of relations
        """
        relation_map = self.get_relations_map()
        relation_list = list()

        for map_parameter_name in relation_map.keys():
            # if compare_two_camel_case_words(map_parameter_name, parameter_name):
            if map_parameter_name.lower()==parameter_name.lower(): # FIXME: mixup of id, ids or other plural things. same for the operation names when creating a relation. GetThing and GetThings
                relation_list.extend(relation_map.get(map_parameter_name,[])) 

        if relation_list:
            non_self_referencing_relations = [
                relation_dict
                for relation_dict in relation_list
                if relation_dict.get('operation_name')!=exclude_operation_name
            ]
            return non_self_referencing_relations
        
        return False



    def find_best_relations_for_operations_parameters(self, operation_name:str) -> List[Dict]:
        resource_node = self.service_node.find_resource_node_by_operation_name(operation_name)
        if not resource_node:
            return False
        relation_map = self
        result = resource_node.find_best_relations_for_operation(operation_name, relation_map)
        return result
            
    def _save_relations_map_to_file(self, relations_map:List[Dict]=None) -> None:
        """Saves the generated relations map to `{BALCONY_RELATIONS_DIR}/{service_name}.json`
        # TODO: remove the parameter as it's not needed
        Args:
            relations_map (List[Dict], optional): _description_. Defaults to None.
        """
        
        directory = BALCONY_RELATIONS_DIR
        service_name = self.service_node.name
        filepath = os.path.join(directory, f"{service_name}.json")
        relations_map = self.get_relations_map()
        with open(filepath, 'w') as file:
            json.dump(relations_map, file, indent=2, default=str)

    def _load_relations_from_file(self):
        directory = BALCONY_RELATIONS_DIR
        service_name = self.service_node.name
        filepath = os.path.join(directory, f"{service_name}.json")
        if not os.path.isfile(filepath):
            return False
        relations_map = None
        try:
            with open(filepath, 'r') as file:
                relations_map = json.load(file)
        except:
            pass
        return relations_map

    def _generate_resource_node_parameters_list(self, resource_nodes):
        _list = []
        for resource_node in resource_nodes:
            for operation_name in resource_node.operation_names:
                required_parameter_names = resource_node.get_required_parameter_names_from_operation_name(operation_name)
                _list.append({
                    'resource_node':resource_node,
                    'operation_name':operation_name,
                    'required_parameter_names':required_parameter_names
                })
        return _list


    def _generate_relation_map(self, **kwargs):
        resource_nodes = self.service_node.get_resource_nodes()
        # Create a list of dicts for each resource_node, operation_name and required_parameter_names
        operations_to_required_parameter_list = self._generate_resource_node_parameters_list(resource_nodes)
        
        # Get every unique required parameter names
        unique_required_shape_names = set()
        for _dict in operations_to_required_parameter_list:
            for r_param_name in _dict.get('required_parameter_names'):
                unique_required_shape_names.add(r_param_name)
        unique_required_shape_names_list = list(unique_required_shape_names)

        all_found_relations = []
        for __dict in operations_to_required_parameter_list:
            # unpack the dict
            resource_node=__dict.get('resource_node')
            operation_name =__dict.get('operation_name')
            required_parameter_names=__dict.get('required_parameter_names')
            
            operation_model = resource_node.get_operation_model(operation_name)
            output_shape = operation_model.output_shape
            
            resource_nodes_extra_relations = resource_node.define_extra_relations()
            if resource_nodes_extra_relations:
                all_found_relations.extend(resource_nodes_extra_relations)
            # for each member shape (attr.) of this operation 
            output_shape_and_target_paths = flatten_shape_to_its_non_collection_shape_and_target_paths(output_shape)
            # try to find it in the unique required parameter list
            for output_shape_and_target_path in output_shape_and_target_paths:
                shape = output_shape_and_target_path.shape
                target_path = output_shape_and_target_path.target_path
                shape_name = get_shape_name(shape)
                prefixed_naked_shape_name = None
                # if it's named only 'id', 'arn'... we append nodename in front of it and add it as an alias.
                if shape_name.lower() in IDENTIFIER_NAMES:
                    prefixed_naked_shape_name = f"{resource_node.name}{shape_name.capitalize()}"
                    all_found_relations.append({
                        'service_name':self.service_node.name,
                        'resource_node_name': resource_node.name,
                        'operation_name': operation_name,
                        'search_shape_name': prefixed_naked_shape_name,
                        'target_shape_name': shape_name,
                        'target_shape_type': shape.type_name,
                        'target_path': target_path,
                        'alias': True
                    })
                for required_shape_name in unique_required_shape_names_list:
                    # TODO: compare lower names
                    _match = icompare_two_camel_case_words(shape_name, required_shape_name)
                    if _match:
                        all_found_relations.append({
                            'service_name':self.service_node.name,
                            'resource_node_name': resource_node.name,
                            'operation_name': operation_name,
                            'search_shape_name': required_shape_name,
                            'target_shape_name': shape_name,
                            'target_shape_type': shape.type_name,
                            'target_path': target_path
                        })
        relations_map = {
            required_shape_name:[]
            for required_shape_name in unique_required_shape_names_list
        }
        def add_to_relations(relation):
            nonlocal relations_map # outer scope access
            alias = relation.get('alias', False)
            parameter_name = relation.get('search_shape_name')
            
            def is_key_in_relations_map(search_key):
                nonlocal relations_map # outer scope access
                map_keys = relations_map.keys()
                for map_key in map_keys:
                    # if compare_two_camel_case_words(map_key, search_key):
                    if map_key.lower()==search_key.lower(): # TODO: decide between these two
                        return map_key
                return False

            is_parameter_in_relation_map = is_key_in_relations_map(parameter_name)
            if is_parameter_in_relation_map:
                selected_parameter_name_key_in_the_map = is_parameter_in_relation_map
                relations_map[selected_parameter_name_key_in_the_map].append(relation)
            else:
                relations_map[parameter_name] = [relation]
        
              
        for found_rel in all_found_relations:
            add_to_relations(found_rel)
          
        return relations_map


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['balcony', 'balcony.custom_nodes']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.24.80,<2.0.0',
 'inflect>=6.0.0,<7.0.0',
 'jmespath>=1.0.1,<2.0.0',
 'rich>=12.5.1,<13.0.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['balcony = balcony.cli:run_app']}

setup_kwargs = {
    'name': 'balcony',
    'version': '0.0.33',
    'description': 'AWS API for humans',
    'long_description': '# balcony\nAWS API for humans\n\n\n## Installation\n\n```bash\npip3 install balcony\n```\n\n## Basic Usage\n\n\n```bash\nbalcony --help\n\n# list all available services\nbalcony aws ls \n\n# list resource nodes of a service\nbalcony aws ls iam\n\n# details about the resource node\nbalcony aws ls iam Policy\n\n# read a Resource Node from AWS API\nbalcony aws read iam Policy\n```',
    'author': 'Oguzhan Yilmaz',
    'author_email': 'oguzhanylmz271@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)




try:
    from .logs import get_logger, get_rich_console
except ImportError:
    from logs import get_logger, get_rich_console
    
from typing import List, Set, Dict, Tuple, Optional, Union
import textwrap

logger = get_logger(__name__)

def generate_py_code_for_customization(error, func_signature, func_ret_type='', help_text=''):
    error_msg = error.message
    resource_node = error.context.get('resource_node')
    service = error.context.get('service')
    operation_name = error.context.get('operation_name')
    err_heading = f"[red bold]Error: {error_msg}[/] for [bold][green]{service}[/].[blue]{resource_node}[/], operation: [yellow]{operation_name}[/][/]"
    file_line = f"[bold]File to edit: [green]balcony/custom_nodes/[/][blue]{service}.py[/][/]"
    class_line = f"class {resource_node}(ResourceNode, service_name='{service}', name='{resource_node}'):"
    # [reverse bold ]Please create a custom class and do something about this issue.[/]
    r = '{r}'
    return textwrap.dedent(f"""
    [yellow bold reverse] ~~ HELP NEEDED! ~~[/]
    {err_heading}
    {help_text}
    
    [underline]{file_line}[/]
    
    try:
        from ..nodes import ResourceNode
        from ..logs import get_logger
    except ImportError:
        from nodes import ResourceNode
        from logs import get_logger
    from typing import List, Set, Dict, Tuple, Optional, Union
    logger = get_logger(__name__)
        
    {class_line}
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            
        def {func_signature}{func_ret_type}:
            result, err = super().{func_signature}
            # manipulate `result` or return your own value
            # logger.debug(result, err)
            return result, None        
    """, ).lstrip()

class Error(Exception):
    def __init__(self, message:str, context:Dict=None) -> None:            
        super().__init__(message) # Call the base class constructor
        self.message = message
        self.context = context
        error_log = self.create_error_log()
        if error_log:
            logger.debug(error_log)

    def create_error_log(self):
        error_msg = self.message

        if error_msg == "failed to generate relations":
            return generate_py_code_for_customization(self, 'get_operations_relations(operation_name)', ' -> Tuple[List[Dict], None]')
        elif error_msg == "failed to choose the best relation":
            return generate_py_code_for_customization(self, 'get_operations_relations(operation_name)', ' -> Tuple[List[Dict], None]')
        elif error_msg == "missing relations for parameter":
            return generate_py_code_for_customization(self, 'get_operations_relations(operation_name)', ' -> Tuple[List[Dict], None]')
        elif error_msg == "failed to generate api parameters":
            pass
        elif error_msg == "failed to generate jmespath selector":
            pass
        elif error_msg == "related resources not found":
            pass
        elif error_msg == "":
            pass
        elif error_msg == "":
            pass
        else:
            return False
